#ifndef BOOTLOADER_H
#define BOOTLOADER_H

#include <cstdint>

#include "Uart.h"

class Bootloader
{
public:
    Bootloader(Uart &uart);

    uint16_t chip_id();
    uint32_t flash_size();
    bool chip_is_supported();

    bool read_memory(uint32_t start_address, uint8_t *mem, uint16_t size);
    bool write_memory(uint32_t start_address, uint8_t *mem, uint16_t size);
    bool full_flash_erase();
    bool go(uint32_t start_address);

    const char *SUPPORTED_DEVICE_NAME = "STM32G03xxx/04xxx";
    const uint32_t MAIN_FLASH_START_ADDRESS = 0x08000000;

    struct ChipData
    {
        uint16_t id;
        uint8_t bootloader_version_major;
        uint8_t bootloader_version_minor;
        uint8_t protocol_version_major;
        uint8_t protocol_version_minor;
        uint8_t option_byte_1;
        uint8_t option_byte_2;
        uint32_t flash_size;
    };

    struct Commands
    {
        uint8_t get;
        uint8_t get_version_and_read_protection_status;
        uint8_t get_id;
        uint8_t read_memory;
        uint8_t go;
        uint8_t write_memory;
        uint8_t erase;
        uint8_t write_protect;
        uint8_t write_unprotect;
        uint8_t readout_protect;
        uint8_t readout_unprotect;
    };

private:
    void syn();
    void get();
    void get_version();
    void get_chip_id();
    void get_flash_memory_size();
    void get_bl_version();

    void send_command(uint8_t command);
    void send_address(uint32_t address);
    bool wait_for_ack();

    const uint32_t FLASH_MEMORY_SIZE_DATA_REGISTER_ADDRESS = 0x1fff75e0;
    const uint32_t BL_VERSION_REGISTER_ADDRESS = 0x1fff1ffe;
    const uint16_t SUPPORTED_DEVICE_ID = 0x466;
    const uint8_t ACK = 0x79;
    const uint8_t NAK = 0x1f;

    const struct Commands default_commands = {
        .get = 0x00,
        .get_version_and_read_protection_status = 0x01,
        .get_id = 0x02,
        .read_memory = 0x11,
        .go = 0x21,
        .write_memory = 0x31,
        .erase = 0x43,
        .write_protect = 0x63,
        .write_unprotect = 0x73,
        .readout_protect = 0x82,
        .readout_unprotect = 0x92,
    };

    ChipData chip = {
        .id = 0,
        .bootloader_version_major = 0,
        .bootloader_version_minor = 0,
        .protocol_version_major = 0,
        .protocol_version_minor = 0,
        .option_byte_1 = 0,
        .option_byte_2 = 0,
        .flash_size = 0,
    };

    Uart &uart;

    Commands commands = default_commands;
    bool is_sync = false;
    bool is_supported = false;
};

#endif
